package io.tswf.groovy.bettergroovy.extensions.instance

import groovy.transform.stc.ClosureParams
import groovy.transform.stc.FirstParam

import java.util.concurrent.ConcurrentHashMap
import java.util.function.Predicate
import java.util.function.Supplier

class AdditionalCollectionMethods {
    static boolean isNullOrEmpty(Collection collection) {
        return collection == null || collection.isEmpty()
    }

    static boolean isNotNullOrEmpty(Collection collection) {
        return !isNullOrEmpty(collection)
    }

    static boolean isNotEmpty(Collection collection) {
        return !collection.isEmpty()
    }

    static <T, R> List<R> map(Collection<T> instance, @ClosureParams(FirstParam.FirstGenericType.class) Closure<R> mapperFunc) {
        return mapToCollection(instance, {new ArrayList<R>()}, mapperFunc)
    }

    static <T> T max(Collection<T> instance, Comparator<T> comparator) {
        return instance.stream()
            .max(comparator)
            .orElse(null)
    }

    static <T> T min(Collection<T> instance, Comparator<T> comparator) {
        return instance.stream()
                .min(comparator)
                .orElse(null)
    }

    static <T> boolean none(Collection<T> instance, Predicate<T> filter) {
        return instance.stream()
            .noneMatch(filter)
    }

    static <T, K> Map<K, List<T>> groupBy(Collection<T> instance, @ClosureParams(FirstParam.FirstGenericType.class) Closure<K> keySupplier) {
        def map = new ConcurrentHashMap<K, List<T>>()
        for (def collectionElement : instance) {
            def key = keySupplier(collectionElement)
            map.computeIfAbsent(key, {new ArrayList<T>()}) << collectionElement
        }

        return map
    }

    static <T, R, C extends Collection<R>> C mapToCollection(Collection<T> instance, Supplier<C> collectionFactory, @ClosureParams(FirstParam.FirstGenericType.class) Closure<R> mapperFunc) {
        def newCollection = collectionFactory.get()
        for (def collectionElement : instance) {
            newCollection.add(mapperFunc(collectionElement))
        }

        return newCollection
    }
}
