package io.tswf.groovy.bettergroovy.extensions.instance

import java.nio.file.Files
import java.nio.file.Path
import java.util.function.Consumer
import java.util.stream.Stream

class AdditionalFileMethods {
    static String readAllAsString(File file, String encoding = "UTF-8") {
        return AdditionalInputStreamMethods.readAllAsString(getInputStream(file), encoding)
    }

    static OutputStream getOutputStream(File file, boolean append = false) {
        return new FileOutputStream(file, append)
    }

    static InputStream getInputStream(File file) {
        return new FileInputStream(file)
    }

    static File subFile(File file, String name) {
        return new File(file, name)
    }

    static void setContent(File file, byte[] bytes) {
        file.newOutputStream().with {
            write(bytes)
            flush()
            close()
        }
    }

    static File createIfNotExists(File file) {
        if (file.exists())
            return

        if (file.name.endsWith("/") || file.name.endsWith("\\") || file.isDirectory()) {
            Files.createDirectories(getAbsoluteCanonicalFile(file).toPath())
        } else {
            def parentDir = file.getParentFile()
            if (parentDir != null) {
                mkdirsUnsafe(parentDir)
            }

            Files.createFile(getAbsoluteCanonicalFile(file).toPath())
        }

        return file
    }

    static File mkdirsUnsafe(File file) {
        if (!file.exists())
            Files.createDirectories(getAbsoluteCanonicalFile(file).toPath())

        return file
    }

    static File getRelativeFromWorkdir(File file) {
        return getRelativeFrom(file, new File("./"))
    }

    static String getNameWithoutExtension(File file) {
        def name = file.name
        def indexOfLastDot = name.lastIndexOf('.')
        if (indexOfLastDot > 0 && indexOfLastDot < name.length() - 2) {
            return name.substring(indexOfLastDot + 1)
        }

        return name
    }

    static boolean hasNameExtension(File file) {
        return !AdditionalStringMethods.isNullOrBlank(getNameExtension(file))
    }

    static String getNameExtension(File file) {
        def name = file.name
        def indexOfLastDot = name.lastIndexOf('.')
        if (indexOfLastDot > 0 && indexOfLastDot < name.length() - 1) {
            return name.substring(0, indexOfLastDot)
        }

        return null
    }

    static File getRelativeFrom(File file, File from) {
        def canonicalToPath = "${getAbsoluteCanonicalFile(from).absolutePath}/".replace('\\', '/')
        def canonicalFilePath = getAbsoluteCanonicalFile(file).absolutePath.replace('\\', '/')
        if (canonicalFilePath.startsWith(canonicalToPath)) {
            return new File(canonicalFilePath.replace(canonicalToPath, ""))
        }

        return new File(getAbsoluteCanonicalFile(from).relativePath(getAbsoluteCanonicalFile(file)).replace("\\", "/"))
    }

    static File getAbsoluteCanonicalFile(File file) {
        def absFle =  file.getAbsoluteFile().getCanonicalFile().getAbsoluteFile()
        if (file.getName().endsWith("/") && !absFle.name.endsWith("\\") && !absFle.name.endsWith("/")) {
            return new File("${absFle.name}/")
        }

        return absFle
    }

    static Stream<File> walkSubFiles(File file, Consumer<Throwable> exceptionsHandlerFun = null) {
        return AdditionalPathMethods.walkSubFiles(file.toPath(), exceptionsHandlerFun).map {getAbsoluteCanonicalFile(it.toFile())}
    }
}
