package io.tswf.groovy.bettergroovy.extensions.instance

import groovy.transform.NamedVariant
import io.tswf.groovy.bettergroovy.extensions.statics.AdditionalSystemStaticMethods

class AdditionalInputStreamMethods {
    static int defaultBufferSize = AdditionalSystemStaticMethods.findGlobalProperty(null, "io.tswf.groovy.bettergroovy.iostreams.buffer.size").map {
        try {
            return Integer.valueOf(it)
        } catch (Throwable e) {
            println "Error while trying to load default buffer size property"
            e.printStackTrace()
        }

        return null
    } .orElse(8096)

    static String readAllAsString(InputStream is, String charset = "UTF-8") {
        return new String(is.readAllBytes(), charset)
    }

    @NamedVariant
    static long copyTo(InputStream is, OutputStream os, boolean closeIn = true, boolean closeOut = true, int bufferSize = defaultBufferSize) {
        byte[] buffer = new byte[bufferSize];

        long totalLen = 0

        int len = is.read(buffer)

        while (len != -1) {
            os.write(buffer, 0, len)
            totalLen += len
            len = is.read(buffer)
        }

        if (closeIn)
            is.close()

        os.flush()

        if (closeOut)
            os.close()

        return totalLen
    }
}
