package io.tswf.groovy.bettergroovy.extensions.instance

import java.nio.file.Files
import java.nio.file.Path
import java.util.function.Consumer
import java.util.stream.Stream
import java.util.stream.StreamSupport

class AdditionalPathMethods {
    static Stream<Path> walkSubFiles(Path file, Consumer<Throwable> exceptionsHandler = null)
    {
        var treeWalker = new _FileTreeWalker(file, exceptionsHandler)

        Spliterator<Path> spliterator =
                Spliterators.spliteratorUnknownSize(new _FileTreeWalkerIterator(treeWalker), Spliterator.DISTINCT)


        return StreamSupport
                .stream(spliterator, false)

    }

    private static class _FileTreeWalker {
        private Consumer<Throwable> _exceptionHandlerFun
        private ArrayDeque<Path> _thisLayer = new ArrayDeque<>()
        private Set<Path> _visitedPaths = new HashSet<>()

        _FileTreeWalker(Path file, Consumer<Throwable> exceptionHandlerFun = null) {
            readFileOrDir(file, true)
            this._exceptionHandlerFun = exceptionHandlerFun
        }

        Path next() {
            var hasNextElement = hasNext()

            if (!hasNextElement)
                return null

            var file = _thisLayer.poll()

            if (file != null)
                readFileOrDir(file, false)

            return file.toAbsolutePath()
        }

        boolean hasNext() {
            return !_thisLayer.isEmpty()
        }

        void readFileOrDir(Path file, boolean selfAdd) {
            if (Files.isDirectory(file)) {
                try {
                    Files.newDirectoryStream(file)
                            .forEach((path) -> {
                                if (_visitedPaths.add(path.toRealPath()))
                                    _thisLayer.push(path)
                            })

                }
                catch (Throwable e) {
                    _handleException(e)
                }
            } else if (selfAdd) {
                if (_visitedPaths.add(file.toRealPath()))
                    _thisLayer.push(file)
            }
        }

        void _handleException(Throwable e) {
            if (_exceptionHandlerFun != null) {
                _exceptionHandlerFun.accept(e)
            }
        }
    }

    private static class _FileTreeWalkerIterator implements Iterator<Path>{
        private _FileTreeWalker _fileTreeWalker

        _FileTreeWalkerIterator(_FileTreeWalker fileTreeWalker)
        {
            this._fileTreeWalker = Objects.requireNonNull(fileTreeWalker, "file tree walker can't be null!")
        }

        @Override
        boolean hasNext()
        {
            return _fileTreeWalker.hasNext()
        }

        @Override
        Path next()
        {
            return _fileTreeWalker.next()
        }
    }
}
