package io.tswf.groovy.bettergroovy.extensions.instance

import groovy.transform.stc.ClosureParams
import groovy.transform.stc.FirstParam

import java.util.stream.Collectors
import java.util.stream.Stream

class AdditionalStreamMethods {
    static <T> Stream<T> stream(T[] arr) {
        return Arrays.stream(arr)
    }

    static <T> List<T> toList(Stream<T> stream) {
        return stream.collect(Collectors.toList())
    }

    static <T> Set<T> toSet(Stream<T> stream) {
        return stream.collect(Collectors.toSet())
    }

    static <T> T firstOrNull(Stream<T> self) {
        return self.findFirst().orElse(null)
    }

    static <T, C extends Comparable> Stream<T> reverseSortedBy(Stream<T> self, @ClosureParams(FirstParam.FirstGenericType) Closure<C> parameterSupplier) {
        return self.sorted(Comparator.comparing {parameterSupplier(it)}.reversed())
    }

    static <T, C extends Comparable> Stream<T> sortedBy(Stream<T> self, @ClosureParams(FirstParam.FirstGenericType) Closure<C> parameterSupplier) {
        return self.sorted(Comparator.comparing {parameterSupplier(it)})
    }

    static <T> Stream<T> joinWith(Stream<T> self, Stream<T> anotherStream) {
        if (anotherStream == null) {
            return self
        } else if (self == null) {
            return anotherStream
        }

        return Stream.concat(self, anotherStream)
    }

    static <T, K> Stream<T> distinctBy(Stream<T> stream, @ClosureParams(FirstParam.FirstGenericType) Closure<K> keySupplier) {
        def keys = new HashSet<K>()
        return stream.filter({keys.add(keySupplier(it))})
    }

    static String joinToString(Stream stream) {
        joinToString(stream, "")
    }

    static String joinToString(Stream stream, String splitter) {
        return stream.map({String.valueOf(it)}).collect(Collectors.joining(splitter))
    }
}
