package io.tswf.groovy.bettergroovy.extensions.instance

import java.util.concurrent.atomic.AtomicInteger
import java.util.function.BiFunction
import java.util.stream.Collectors

class AdditionalStringMethods {
    static String trimLines(String str) {
        def lineSep = getLineSeparator(str)

        str.split(lineSep)
                .stream()
                .map({it.trim()})
                .collect(Collectors.joining(lineSep))
    }

    static boolean isMultiline(String str) {
        return getLineSeparator(str) != null
    }

    static String stripAll(String str) {
        return str.stripTrailing().stripLeading()
    }

    static boolean isNullOrEmpty(String str) {
        return str == null || str.isEmpty()
    }

    static boolean isNullOrBlank(String str) {
        return str == null || stripAll(str).isEmpty()
    }

    static int getLinesCount(String str) {
        def lineSep = getLineSeparator(str)
        if (lineSep != null) {
            return str.count(lineSep) + 1
        }

        return 1
    }

    static String removeLinesIf(String str, BiFunction<String, Integer, Boolean> filter) {
        def sep = getLineSeparator(str)
        if (sep == null) {
            return str
        }

        def nextStrId = new AtomicInteger()
        return Arrays.stream(str.split(sep))
                .filter {!filter(it, nextStrId.getAndIncrement())}
                .collect(Collectors.joining(sep))
    }

    static String getLineSeparator(String str) {
        def lf = "\n"
        def cr = "\r"
        def crlf = "$cr$lf"
        if (str.contains(crlf)) {
            return crlf
        } else if (str.contains(lf)) {
            return lf
        } else if (str.contains(cr)) {
            return cr
        }

        return null
    }
}
