package io.tswf.groovy.bettergroovy.extensions.statics

import java.util.concurrent.ConcurrentHashMap;

class AdditionalSystemStaticMethods {

    static Set<URL> globalConfigurationsLocations = Collections.<URL>newSetFromMap(new ConcurrentHashMap<>())

    static void registerGlobalConfiguration(System sys, URL url) {
        globalConfigurationsLocations.add(url)
    }

    static void registerGlobalConfiguration(System sys, File file) {
        registerGlobalConfiguration(sys, file.toPath().toUri().toURL())
    }

    static String getGlobalProperty(System sys, String propertyName) {
        return findGlobalProperty(sys, propertyName).orElseThrow(() -> new NoSuchElementException("Can't find property '$propertyName' in application global configuration!"))
    }

    static String getGlobalPropertyOrNull(System sys, String propertyName) {
        return findGlobalProperty(sys, propertyName).orElse(null)
    }

    static Optional<String> findGlobalProperty(System sys, String propertyName) {
        var env = System.getenv(propertyName.replace(".", "_").replace("-", "_").toUpperCase())
        if (env != null) {
            return Optional.of(env)
        }

        var dProperty = System.getProperty(propertyName)

        if (dProperty != null) {
            return Optional.of(dProperty)
        }

        for (def globalConfigurationLocation : globalConfigurationsLocations) {
            def properties = new Properties()
            properties.load(new InputStreamReader(globalConfigurationLocation.openStream(), "UTF-8"))
            if (properties.containsKey(propertyName)) {
                return Optional.ofNullable(properties.getProperty(propertyName))
            }
        }

        return Optional.empty()
    }
}
