package io.tswf.groovy.bettergroovy.transforms.annotationgstrings


import org.apache.groovy.ast.tools.ExpressionUtils
import org.codehaus.groovy.ast.ASTNode
import org.codehaus.groovy.ast.AnnotationNode
import org.codehaus.groovy.ast.ClassHelper
import org.codehaus.groovy.ast.ClassNode
import org.codehaus.groovy.ast.expr.ConstantExpression
import org.codehaus.groovy.ast.expr.Expression
import org.codehaus.groovy.ast.expr.GStringExpression
import org.codehaus.groovy.control.CompilePhase
import org.codehaus.groovy.control.SourceUnit
import org.codehaus.groovy.transform.AbstractASTTransformation
import org.codehaus.groovy.transform.GroovyASTTransformation
import org.codehaus.groovy.transform.GroovyASTTransformationClass
import org.codehaus.groovy.vmplugin.VMPluginFactory

import java.lang.annotation.Retention
import java.lang.annotation.RetentionPolicy

@GroovyASTTransformationClass("io.tswf.groovy.bettergroovy.transforms.annotationgstrings.GStringsInAnnotationsTransformation")
@Retention(RetentionPolicy.SOURCE)
@interface GStringsInAnnotations {

}

interface GStringMultilineConversions{
    static final String TRIM_MULTILINE_INDENTS = "<TRIM_INDENTS>fdjfhk433rhfeduofrj3reifd"
    static final String TRIM_MULTILINE_TO_FIRST = "<TRIM_TO_FIRST>fdjfhk433rhfeduofrj3reifd"
}

@GroovyASTTransformation(phase = CompilePhase.CONVERSION)
class GStringsInAnnotationsTransformation extends AbstractASTTransformation
{
    static File file = new File("C:\\Users\\theswayfarer\\Desktop\\членсоситупаясука.txt")
    static BufferedWriter w = file.newWriter(true)

    static void println(String str) {
        w.write(str)
        w.newLine()
        w.flush()
    }

    GStringsInAnnotationsTransformation() {
        println "Created a $this instance!"
    }

    @Override
    void visit(ASTNode[] nodes, SourceUnit source) {
        init(nodes, source)

        println "Try to tranform nodes $nodes"

        def cn = nodes.find {it instanceof ClassNode} as ClassNode
        println "Transforming annotaitons on class $cn.name"

        cn.annotations.forEach {transformGStringsInAnnotation(it)}

        cn.allDeclaredMethods.forEach {mn ->
            mn.annotations.forEach {transformGStringsInAnnotation(it)}

            mn.parameters.each {pn ->
                pn.annotations.forEach {transformGStringsInAnnotation(it)}
            }
        }

        cn.declaredConstructors.forEach {mn ->
            mn.annotations.forEach {transformGStringsInAnnotation(it)}

            mn.parameters.each {pn ->
                pn.annotations.forEach {transformGStringsInAnnotation(it)}
            }
        }

        cn.fields.forEach {mn ->
            mn.annotations.forEach {transformGStringsInAnnotation(it)}
        }
    }

    void transformGStringsInAnnotation(AnnotationNode annotation) {
        def members = new HashMap<>(annotation.members)
        for (def memberEntry : members) {
            def memberValueExpr = memberEntry.value
            if (memberValueExpr instanceof GStringExpression) {
                def memberGStringValue = memberValueExpr as GStringExpression
                annotation.members[memberEntry.key] = transformGStringToContant(memberGStringValue)
            } else {
                println "Skip annotation's $annotation.classNode.name $memberEntry.key parameter: $memberValueExpr of type: ${memberValueExpr.class.name}"
            }
        }

        VMPluginFactory.getPlugin().configureAnnotation(annotation)
    }

    ConstantExpression transformGStringToContant(GStringExpression gstring) {
        def constValues = new ArrayList<Expression>()
        println "Transforming gstring expression: ${gstring}"

        for (def value : gstring.values) {
            if (value instanceof GStringExpression) {
                println "Transforming value: $value"
                constValues.add transformGStringToContant(value as GStringExpression)
            } else {
                println "Transforming param $value of $gstring"
                def str = ExpressionUtils.transformInlineConstants(value, ClassHelper.make(String))
                while (str instanceof GStringExpression) {
                    str = transformGStringToContant(str as GStringExpression)
                }
                constValues.add str
            }
        }
        def iterValues = constValues.iterator()

        StringBuilder buffer = new StringBuilder();
        for (def string : gstring.strings) {
            buffer.append(string.getValue())
            if (iterValues.hasNext()) {
                buffer.append((iterValues.next() as ConstantExpression).value)
            }
        }

        def resultString = (String) buffer.toString()

        def lineSep = resultString.getLineSeparator()
        if (lineSep != null) {
            def firstLineEmpty = resultString.lines().limit(1).anyMatch {it.stripLeading().stripTrailing().isEmpty()}

            if (resultString.contains(GStringMultilineConversions.TRIM_MULTILINE_INDENTS)) {
                resultString = resultString
                        .replace(GStringMultilineConversions.TRIM_MULTILINE_INDENTS, "")
                        .trimLines()
            } else if (firstLineEmpty || resultString.contains(GStringMultilineConversions.TRIM_MULTILINE_TO_FIRST)) {
                resultString = resultString
                        .replace(GStringMultilineConversions.TRIM_MULTILINE_TO_FIRST, "")
                        .stripIndent(true)
            }

            resultString = resultString.removeLinesIf{ ln, num -> num == 0 && ln.isEmpty()}

        }

        return new ConstantExpression(resultString)
    }

}
