package io.tswf.groovy.bettergroovy.transforms.asyncs

import org.codehaus.groovy.ast.ASTNode
import org.codehaus.groovy.ast.AnnotationNode
import org.codehaus.groovy.ast.ClassHelper
import org.codehaus.groovy.ast.MethodNode
import org.codehaus.groovy.ast.stmt.BlockStatement
import org.codehaus.groovy.control.CompilePhase
import org.codehaus.groovy.control.SourceUnit
import org.codehaus.groovy.transform.AbstractASTTransformation
import org.codehaus.groovy.transform.GroovyASTTransformation
import org.codehaus.groovy.transform.GroovyASTTransformationClass

import java.lang.annotation.ElementType
import java.lang.annotation.Retention
import java.lang.annotation.RetentionPolicy
import java.lang.annotation.Target
import java.util.concurrent.CompletableFuture

import static org.codehaus.groovy.ast.tools.GeneralUtils.*

@Target(ElementType.METHOD)
@Retention(RetentionPolicy.SOURCE)
@GroovyASTTransformationClass(classes = AsyncMethodsTransform)
@interface AsyncCall {
    boolean await() default true
}

@GroovyASTTransformation(phase = CompilePhase.SEMANTIC_ANALYSIS)
class AsyncMethodsTransform extends AbstractASTTransformation {

    @Override
    void visit(ASTNode[] nodes, SourceUnit source) {
        init(nodes, source)
        def annotation = nodes[0]
        def owner = nodes[1]
        if (annotation instanceof AnnotationNode && owner instanceof MethodNode) {
            def await = annotation?.members["await"]?.text != "false"
            if (!await && !owner.isVoidMethod()) {
                addError("Can't make non-void async method non-awaitable! See ${CompletableFuture.class.name} instead!", owner)
            } else {
                def futureStaticFactoryMethodName = owner.isVoidMethod() ? "runAsync" : "supplyAsync"
                def methodCode = owner.code

                if (methodCode instanceof BlockStatement) {
                    def closure = owner.extractCodeToClosure()
                    def futureExpression = callX(ClassHelper.make(CompletableFuture), futureStaticFactoryMethodName, args(closure))
                    if (await) {
                        futureExpression = callX(futureExpression, "get")
                    }

                    owner.code = block(
                            stmt(futureExpression)
                    )
                }
            }
        }
    }
}
