package io.tswf.groovy.bettergroovy.transforms.directnamedargs

import groovy.transform.CompileStatic
import groovy.transform.Generated
import groovy.transform.NamedParam
import groovy.transform.NamedVariant
import org.codehaus.groovy.ast.*
import org.codehaus.groovy.ast.expr.*
import org.codehaus.groovy.classgen.asm.sc.StaticTypesTypeChooser
import org.codehaus.groovy.control.CompilePhase
import org.codehaus.groovy.control.SourceUnit
import org.codehaus.groovy.transform.GroovyASTTransformation
import org.codehaus.groovy.transform.GroovyASTTransformationClass
import org.codehaus.groovy.transform.sc.StaticCompilationMetadataKeys

import java.lang.annotation.ElementType
import java.lang.annotation.Retention
import java.lang.annotation.RetentionPolicy
import java.lang.annotation.Target

@Target(ElementType.TYPE)
@Retention(RetentionPolicy.SOURCE)
@GroovyASTTransformationClass("io.tswf.groovy.bettergroovy.transforms.directnamedargs.DirectNamedArgsTransform")
@interface DirectNamedArgsCall {

}

@GroovyASTTransformation(phase = CompilePhase.INSTRUCTION_SELECTION)
class DirectNamedArgsTransform extends MethodCallTransformation {

    @Override
    protected GroovyCodeVisitor getTransformer(ASTNode[] nodes, SourceUnit sourceUnit) {
        def target = nodes[1]
        return new MethodCallVisitor(sourceUnit, target instanceof ClassNode ? target as ClassNode : null)
    }
}

@CompileStatic
class MethodCallVisitor extends ClassCodeExpressionTransformer {

    private final StaticTypesTypeChooser typeChooser = new StaticTypesTypeChooser()

    SourceUnit sourceUnit
    ClassNode classNode
    MethodCallVisitor(SourceUnit sourceUnit, ClassNode node) {
        this.sourceUnit = sourceUnit
        this.classNode = node
    }

    @Override
    Expression transform(final Expression exp) {
        if (exp instanceof MethodCallExpression) {
            return transformMethodInvocation(exp)
        } else if (exp instanceof StaticMethodCallExpression) {
            return transformMethodInvocation(exp)
        }

        return super.transform(exp)
    }

    Expression transformMethodInvocation(Expression callExpression) {
        if (classNode == null) {
            return callExpression
        }

        def imethodCall = callExpression as MethodCall
        def isMethodCallStatic = callExpression instanceof StaticMethodCallExpression

        def callArgsExpr = imethodCall.arguments
        if (callArgsExpr instanceof TupleExpression) {
            def callArgsTuple = callArgsExpr as TupleExpression

            if (callArgsTuple?.expressions?.size() == 1 && callArgsTuple.expressions[0] instanceof NamedArgumentListExpression) {
                def callArgsList = callArgsTuple.expressions[0] as NamedArgumentListExpression

                if (callArgsList?.mapEntryExpressions?.isNotEmpty()) {
                    def isAllEntriesKeysIsStringConstants = callArgsList.mapEntryExpressions.any {arg ->
                        if (!(arg.keyExpression instanceof ConstantExpression))
                            false
                        else if (!((arg.keyExpression as ConstantExpression).value instanceof String))
                            false
                        else
                            true
                    }

                    if (isAllEntriesKeysIsStringConstants) {
                        def dynamic = null

                        if (callExpression instanceof MethodCallExpression) {
                            dynamic = callExpression.objectExpression.getNodeMetaData(StaticCompilationMetadataKeys.RECEIVER_OF_DYNAMIC_PROPERTY)
                        }

                        if (dynamic == null) {
                            ClassNode mapMethodOwnerClass = (ClassNode) null

                            if (callExpression instanceof MethodCallExpression) {
                                mapMethodOwnerClass = typeChooser.resolveType(callExpression.objectExpression, classNode)
                            } else if (callExpression instanceof StaticMethodCallExpression) {
                                mapMethodOwnerClass = callExpression.ownerType
                            } else {
                                return callExpression
                            }

                            mapMethodOwnerClass = unpackStaticClassReciever(mapMethodOwnerClass)

                            def mapMethod= mapMethodOwnerClass.allDeclaredMethods.find {it.name == imethodCall.methodAsString && it.parameters.length == 1 && it.parameters[0].type.name == Map.class.name}
                            def isMapMethodGenerated = mapMethod != null && mapMethod?.annotations?.any {it.classNode.name == Generated.class.name}

                            //TODO: а в груви 3 тоже такая аннотация7
                            def isMapMethodParamAnnotatedWithNamedParams = mapMethod?.parameters?.any {param -> param.annotations.any {it.classNode.name == NamedParam.class.name}}

                            def dbgSameNamedMethods = mapMethodOwnerClass.allDeclaredMethods
                                .stream()
                                    .filter {it.name == imethodCall.methodAsString}
                                    .map {"${it.declaringClass.name}:${it.name}(${it.parameters.stream().map{p -> p.type.name}.joinToString(",")})"}
                                    .toList()

                            if (isMapMethodGenerated && isMapMethodParamAnnotatedWithNamedParams) {
                                def nonMapMethod = mapMethodOwnerClass.getDeclaredMethods(mapMethod.name)
                                    .stream()
                                        .filter {method -> method.annotations.any {it.classNode.name == NamedVariant.class.name}}
                                        .filter {!it.annotations.any {a -> a.classNode.name == Generated.class.name}}
                                        .sorted(Comparator.comparingInt { MethodNode it -> it.parameters.size()}.reversed())
                                        .findFirst().orNull()

                                if (nonMapMethod != null) {
                                    def nonMapMethodParams = nonMapMethod.parameters
                                    def callArgsNameToExprMap = new HashMap<String, MapEntryExpression>()

                                    for (def callArgExpr : callArgsList.mapEntryExpressions) {
                                        def mapEntryExprArg = callArgExpr as MapEntryExpression
                                        def constKey = (mapEntryExprArg.keyExpression as ConstantExpression)
                                        def constKeyStringValue = constKey.value as String
                                        callArgsNameToExprMap[constKeyStringValue] = callArgExpr
                                    }

                                    def newInvocationParams = new ArrayList<Expression>()

                                    for (def nonMapMethodParam : nonMapMethodParams) {
                                        def argValueFromMap = callArgsNameToExprMap[nonMapMethodParam.name]
                                        if (argValueFromMap != null) {
                                            newInvocationParams.add(argValueFromMap.valueExpression)
                                            continue
                                        }

                                        if (nonMapMethodParam.hasInitialExpression()) {
                                            def defaultArgValue = nonMapMethodParam.initialExpression
                                            if (defaultArgValue != null) {
                                                newInvocationParams.add(defaultArgValue)
                                                continue
                                            }
                                        }

                                        addError("Can't resolve required arg: ${nonMapMethodParam.name} of ${mapMethodOwnerClass.name}.${nonMapMethod.name}", callExpression)
                                    }

                                    if (callExpression instanceof MethodCallExpression) {
                                        callExpression.methodTarget = nonMapMethod
                                        callExpression.arguments = new ArgumentListExpression(newInvocationParams)
                                        return callExpression
                                    } else if (callExpression instanceof StaticMethodCallExpression) {
                                        return new StaticMethodCallExpression(callExpression.ownerType, callExpression.method, new ArgumentListExpression(newInvocationParams))
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }

        return callExpression
    }

    // Понятия не имею почему, но если добавлять этот AST в груви через конфиг для компилятора у градла, то овнеры у статических вызовов оборачиваются в java.lang.Class
    protected ClassNode unpackStaticClassReciever(ClassNode cn) {
        if (cn == null || cn.name != "java.lang.Class") {
            return cn
        }

        if (cn?.genericsTypes?.length == 1) {
            return cn.genericsTypes[0].type
        }

        return cn
    }

    @Override
    protected SourceUnit getSourceUnit() {
        return sourceUnit
    }
}
