package io.tswf.groovy.bettergroovy.transforms.extensions

import org.codehaus.groovy.ast.ASTNode
import org.codehaus.groovy.ast.AnnotationNode
import org.codehaus.groovy.ast.ClassNode
import org.codehaus.groovy.control.CompilePhase
import org.codehaus.groovy.control.SourceUnit
import org.codehaus.groovy.transform.AbstractASTTransformation
import org.codehaus.groovy.transform.GroovyASTTransformation
import org.codehaus.groovy.transform.GroovyASTTransformationClass

import java.lang.annotation.ElementType
import java.lang.annotation.Retention
import java.lang.annotation.RetentionPolicy
import java.lang.annotation.Target

@GroovyASTTransformation(phase = CompilePhase.SEMANTIC_ANALYSIS)
class ExtensionContainerSearchPseudoTransformation extends AbstractASTTransformation{
    @Override
    void visit(ASTNode[] nodes, SourceUnit source) {
        init(nodes, source)
        def annotation = nodes[0]
        def owner = nodes[1]
        if (owner instanceof ClassNode && annotation instanceof AnnotationNode) {
            def isStaticExtensions = annotation.members["staticExtensions"]?.text == "true"

            if (isStaticExtensions) {
                ExtensionMethodRegistry.foundStaticExtensionClassNodes.add(owner)
            } else {
                ExtensionMethodRegistry.foundInstanceExtensionClassNodes.add(owner)
            }
        }
    }
}

@GroovyASTTransformationClass(classes = ExtensionContainerSearchPseudoTransformation)
@Target(ElementType.TYPE)
@Retention(RetentionPolicy.RUNTIME)
@interface ExtensionMethodsContainer {
    boolean staticExtensions() default false
}