package io.tswf.groovy.bettergroovy.transforms.extensions

import groovy.transform.CompilationUnitAware
import org.codehaus.groovy.ast.ASTNode
import org.codehaus.groovy.ast.AnnotatedNode
import org.codehaus.groovy.ast.ClassNode
import org.codehaus.groovy.control.CompilationUnit
import org.codehaus.groovy.control.CompilePhase
import org.codehaus.groovy.control.SourceUnit
import org.codehaus.groovy.transform.ASTTransformation
import org.codehaus.groovy.transform.GroovyASTTransformation
import org.codehaus.groovy.transform.GroovyASTTransformationClass
import org.codehaus.groovy.transform.stc.StaticTypeCheckingVisitor

import java.lang.annotation.ElementType
import java.lang.annotation.Retention
import java.lang.annotation.RetentionPolicy
import java.lang.annotation.Target

@GroovyASTTransformationClass(classes = SilentStaticTypeCheckerTransform)
@Target(ElementType.TYPE)
@Retention(RetentionPolicy.SOURCE)
@interface SilentStaticTypeChecker {

}

class MyStaticTypeCheckingVisitor extends StaticTypeCheckingVisitor {

    MyStaticTypeCheckingVisitor(SourceUnit source, ClassNode classNode) {
        super(source, classNode)
    }

    @Override
    void addError(String msg, ASTNode node) {
        // Nope!
    }
}

@GroovyASTTransformation(phase = CompilePhase.CANONICALIZATION)
class SilentStaticTypeCheckerTransform implements ASTTransformation, CompilationUnitAware{
    protected CompilationUnit compilationUnit

    @Override
    void setCompilationUnit(CompilationUnit unit) {
        compilationUnit = unit
    }

    @Override
    void visit(ASTNode[] nodes, SourceUnit source) {
        AnnotatedNode node = (AnnotatedNode) nodes[1]
        ClassNode classNode = (ClassNode) node
        def visitor = new MyStaticTypeCheckingVisitor(source, classNode)
        visitor.setCompilationUnit(compilationUnit)
        visitor.initialize()
        visitor.visitClass(classNode)
        visitor.performSecondPass()
    }
}
