package io.tswf.groovy.bettergroovy.transforms.namedvariantdefaults

import groovy.transform.NamedVariant
import org.codehaus.groovy.ast.ASTNode
import org.codehaus.groovy.ast.ClassHelper
import org.codehaus.groovy.ast.MethodNode
import org.codehaus.groovy.ast.expr.*
import org.codehaus.groovy.ast.stmt.BlockStatement
import org.codehaus.groovy.ast.stmt.ExpressionStatement
import org.codehaus.groovy.control.SourceUnit
import org.codehaus.groovy.transform.AbstractASTTransformation
import org.codehaus.groovy.transform.GroovyASTTransformation

@GroovyASTTransformation
class NamedVariantDefaultValuesTransformation extends AbstractASTTransformation{

    @Override
    void visit(ASTNode[] nodes, SourceUnit source) {
        this.init(nodes, source)

        def node = nodes.find {it instanceof MethodNode}
        if (node != null && node instanceof MethodNode && node.annotations.any {a -> a.classNode.name == NamedVariant.class.name}) {
            def originalMethodNode = node as MethodNode

            def mapMethodNode = originalMethodNode.declaringClass.getMethods(originalMethodNode.name)
                    .find {it.parameters.length == 1 && it.parameters[0].type.name == Map.class.name  && it.annotations.any {a -> a.classNode.name == "groovy.transform.Generated"} }

            if (mapMethodNode != null && mapMethodNode.code instanceof BlockStatement) {
                def methodCode = mapMethodNode.code as BlockStatement
                if (methodCode.statements[1] instanceof ExpressionStatement) {
                    def exStatement = methodCode.statements[1] as ExpressionStatement
                    def originalMethodCall = exStatement.expression as MethodCallExpression
                    def newArgs = new ArgumentListExpression()

                    def namedArgsVar = mapMethodNode.parameters[0]
                    
                    for (def arg : originalMethodNode.parameters) {
                        newArgs.addExpression new TernaryExpression(
                            new BooleanExpression(
                                new MethodCallExpression(new VariableExpression(namedArgsVar), "containsKey", new ArgumentListExpression(new ConstantExpression(arg.name)))
                            ),
                            new MethodCallExpression(new VariableExpression(namedArgsVar), "get", new ArgumentListExpression(new ConstantExpression(arg.name))),
                                arg.hasInitialExpression() ? arg.initialExpression : new StaticMethodCallExpression(ClassHelper.make(NamedVariantDefaultValuesThrower.class), "missingNamedParameter", new ArgumentListExpression(new ConstantExpression(originalMethodNode.name), new ConstantExpression(arg.name)))
                        )
                        exStatement.setExpression(new MethodCallExpression(originalMethodCall.objectExpression, originalMethodCall.method, newArgs))
                    }
                }
            }
        }
    }
}
